@@ -21,12 +21,13 @@
 
 #include "php_phongo.h"
 #include "phongo_error.h"
+#include "Symbol_arginfo.h"
 
 zend_class_entry* php_phongo_symbol_ce;
 
 /* Initialize the object and return whether it was successful. An exception will
  * be thrown on error. */
-static bool php_phongo_symbol_init(php_phongo_symbol_t* intern, const char* symbol, size_t symbol_len) /* {{{ */
+static bool php_phongo_symbol_init(php_phongo_symbol_t* intern, const char* symbol, size_t symbol_len)
 {
 	if (strlen(symbol) != (size_t) symbol_len) {
 		phongo_throw_exception(PHONGO_ERROR_INVALID_ARGUMENT, "Symbol cannot contain null bytes");
@@ -37,11 +38,11 @@ static bool php_phongo_symbol_init(php_phongo_symbol_t* intern, const char* symb
 	intern->symbol_len = symbol_len;
 
 	return true;
-} /* }}} */
+}
 
 /* Initialize the object from a HashTable and return whether it was successful.
  * An exception will be thrown on error. */
-static bool php_phongo_symbol_init_from_hash(php_phongo_symbol_t* intern, HashTable* props) /* {{{ */
+static bool php_phongo_symbol_init_from_hash(php_phongo_symbol_t* intern, HashTable* props)
 {
 	zval* symbol;
 
@@ -51,9 +52,9 @@ static bool php_phongo_symbol_init_from_hash(php_phongo_symbol_t* intern, HashTa
 
 	phongo_throw_exception(PHONGO_ERROR_INVALID_ARGUMENT, "%s initialization requires \"symbol\" string field", ZSTR_VAL(php_phongo_symbol_ce->name));
 	return false;
-} /* }}} */
+}
 
-HashTable* php_phongo_symbol_get_properties_hash(phongo_compat_object_handler_type* object, bool is_temp) /* {{{ */
+HashTable* php_phongo_symbol_get_properties_hash(phongo_compat_object_handler_type* object, bool is_temp)
 {
 	php_phongo_symbol_t* intern;
 	HashTable*           props;
@@ -74,52 +75,54 @@ HashTable* php_phongo_symbol_get_properties_hash(phongo_compat_object_handler_ty
 	}
 
 	return props;
-} /* }}} */
+}
 
-/* {{{ proto string MongoDB\BSON\Symbol::__toString()
-   Return the Symbol's symbol string. */
-static PHP_METHOD(Symbol, __toString)
+PHONGO_DISABLED_CONSTRUCTOR(MongoDB_BSON_Symbol)
+
+/* Return the Symbol's symbol string. */
+static PHP_METHOD(MongoDB_BSON_Symbol, __toString)
 {
-	zend_error_handling  error_handling;
 	php_phongo_symbol_t* intern;
 
-	zend_replace_error_handling(EH_THROW, phongo_exception_from_phongo_domain(PHONGO_ERROR_INVALID_ARGUMENT), &error_handling);
-	if (zend_parse_parameters_none() == FAILURE) {
-		zend_restore_error_handling(&error_handling);
-		return;
-	}
-	zend_restore_error_handling(&error_handling);
+	PHONGO_PARSE_PARAMETERS_NONE();
 
 	intern = Z_SYMBOL_OBJ_P(getThis());
 
 	RETURN_STRINGL(intern->symbol, intern->symbol_len);
-} /* }}} */
+}
 
-/* {{{ proto array MongoDB\BSON\Symbol::jsonSerialize()
-*/
-static PHP_METHOD(Symbol, jsonSerialize)
+static PHP_METHOD(MongoDB_BSON_Symbol, __set_state)
+{
+	php_phongo_symbol_t* intern;
+	HashTable*           props;
+	zval*                array;
+
+	PHONGO_PARSE_PARAMETERS_START(1, 1)
+	Z_PARAM_ARRAY(array)
+	PHONGO_PARSE_PARAMETERS_END();
+
+	object_init_ex(return_value, php_phongo_symbol_ce);
+
+	intern = Z_SYMBOL_OBJ_P(return_value);
+	props  = Z_ARRVAL_P(array);
+
+	php_phongo_symbol_init_from_hash(intern, props);
+}
+
+static PHP_METHOD(MongoDB_BSON_Symbol, jsonSerialize)
 {
-	zend_error_handling  error_handling;
 	php_phongo_symbol_t* intern;
 
-	zend_replace_error_handling(EH_THROW, phongo_exception_from_phongo_domain(PHONGO_ERROR_INVALID_ARGUMENT), &error_handling);
-	if (zend_parse_parameters_none() == FAILURE) {
-		zend_restore_error_handling(&error_handling);
-		return;
-	}
-	zend_restore_error_handling(&error_handling);
+	PHONGO_PARSE_PARAMETERS_NONE();
 
 	intern = Z_SYMBOL_OBJ_P(getThis());
 
 	array_init_size(return_value, 1);
 	ADD_ASSOC_STRINGL(return_value, "$symbol", intern->symbol, intern->symbol_len);
-} /* }}} */
+}
 
-/* {{{ proto string MongoDB\BSON\Symbol::serialize()
-*/
-static PHP_METHOD(Symbol, serialize)
+static PHP_METHOD(MongoDB_BSON_Symbol, serialize)
 {
-	zend_error_handling  error_handling;
 	php_phongo_symbol_t* intern;
 	zval                 retval;
 	php_serialize_data_t var_hash;
@@ -127,12 +130,7 @@ static PHP_METHOD(Symbol, serialize)
 
 	intern = Z_SYMBOL_OBJ_P(getThis());
 
-	zend_replace_error_handling(EH_THROW, phongo_exception_from_phongo_domain(PHONGO_ERROR_INVALID_ARGUMENT), &error_handling);
-	if (zend_parse_parameters_none() == FAILURE) {
-		zend_restore_error_handling(&error_handling);
-		return;
-	}
-	zend_restore_error_handling(&error_handling);
+	PHONGO_PARSE_PARAMETERS_NONE();
 
 	array_init_size(&retval, 1);
 	ADD_ASSOC_STRINGL(&retval, "symbol", intern->symbol, intern->symbol_len);
@@ -146,13 +144,10 @@ static PHP_METHOD(Symbol, serialize)
 
 	smart_str_free(&buf);
 	zval_ptr_dtor(&retval);
-} /* }}} */
+}
 
-/* {{{ proto void MongoDB\BSON\Symbol::unserialize(string $serialized)
-*/
-static PHP_METHOD(Symbol, unserialize)
+static PHP_METHOD(MongoDB_BSON_Symbol, unserialize)
 {
-	zend_error_handling    error_handling;
 	php_phongo_symbol_t*   intern;
 	char*                  serialized;
 	size_t                 serialized_len;
@@ -161,12 +156,9 @@ static PHP_METHOD(Symbol, unserialize)
 
 	intern = Z_SYMBOL_OBJ_P(getThis());
 
-	zend_replace_error_handling(EH_THROW, phongo_exception_from_phongo_domain(PHONGO_ERROR_INVALID_ARGUMENT), &error_handling);
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "s", &serialized, &serialized_len) == FAILURE) {
-		zend_restore_error_handling(&error_handling);
-		return;
-	}
-	zend_restore_error_handling(&error_handling);
+	PHONGO_PARSE_PARAMETERS_START(1, 1)
+	Z_PARAM_STRING(serialized, serialized_len)
+	PHONGO_PARSE_PARAMETERS_END();
 
 	PHP_VAR_UNSERIALIZE_INIT(var_hash);
 	if (!php_var_unserialize(&props, (const unsigned char**) &serialized, (unsigned char*) serialized + serialized_len, &var_hash)) {
@@ -180,20 +172,16 @@ static PHP_METHOD(Symbol, unserialize)
 
 	php_phongo_symbol_init_from_hash(intern, HASH_OF(&props));
 	zval_ptr_dtor(&props);
-} /* }}} */
+}
 
-/* {{{ proto array MongoDB\Driver\Symbol::__serialize()
-*/
-static PHP_METHOD(Symbol, __serialize)
+static PHP_METHOD(MongoDB_BSON_Symbol, __serialize)
 {
 	PHONGO_PARSE_PARAMETERS_NONE();
 
 	ZVAL_ARR(return_value, php_phongo_symbol_get_properties_hash(PHONGO_COMPAT_OBJ_P(getThis()), true));
-} /* }}} */
+}
 
-/* {{{ proto void MongoDB\Driver\Symbol::__unserialize(array $data)
-*/
-static PHP_METHOD(Symbol, __unserialize)
+static PHP_METHOD(MongoDB_BSON_Symbol, __unserialize)
 {
 	zval* data;
 
@@ -202,45 +190,12 @@ static PHP_METHOD(Symbol, __unserialize)
 	PHONGO_PARSE_PARAMETERS_END();
 
 	php_phongo_symbol_init_from_hash(Z_SYMBOL_OBJ_P(getThis()), Z_ARRVAL_P(data));
-} /* }}} */
+}
 
-/* {{{ MongoDB\BSON\Symbol function entries */
-/* clang-format off */
-ZEND_BEGIN_ARG_WITH_RETURN_TYPE_INFO_EX(ai_Symbol___toString, 0, 0, IS_STRING, 0)
-ZEND_END_ARG_INFO()
-
-ZEND_BEGIN_ARG_INFO_EX(ai_Symbol___unserialize, 0, 0, 1)
-	ZEND_ARG_ARRAY_INFO(0, data, 0)
-ZEND_END_ARG_INFO()
-
-ZEND_BEGIN_ARG_WITH_TENTATIVE_RETURN_TYPE_INFO_EX(ai_Symbol_jsonSerialize, 0, 0, IS_ARRAY, 0)
-ZEND_END_ARG_INFO()
-
-ZEND_BEGIN_ARG_INFO_EX(ai_Symbol_unserialize, 0, 0, 1)
-	ZEND_ARG_INFO(0, serialized)
-ZEND_END_ARG_INFO()
-
-ZEND_BEGIN_ARG_INFO_EX(ai_Symbol_void, 0, 0, 0)
-ZEND_END_ARG_INFO()
-
-static zend_function_entry php_phongo_symbol_me[] = {
-	/* __set_state intentionally missing */
-	PHP_ME(Symbol, __serialize, ai_Symbol_void, ZEND_ACC_PUBLIC | ZEND_ACC_FINAL)
-	PHP_ME(Symbol, __toString, ai_Symbol___toString, ZEND_ACC_PUBLIC | ZEND_ACC_FINAL)
-	PHP_ME(Symbol, __unserialize, ai_Symbol___unserialize, ZEND_ACC_PUBLIC | ZEND_ACC_FINAL)
-	PHP_ME(Symbol, jsonSerialize, ai_Symbol_jsonSerialize, ZEND_ACC_PUBLIC | ZEND_ACC_FINAL)
-	PHP_ME(Symbol, serialize, ai_Symbol_void, ZEND_ACC_PUBLIC | ZEND_ACC_FINAL)
-	PHP_ME(Symbol, unserialize, ai_Symbol_unserialize, ZEND_ACC_PUBLIC | ZEND_ACC_FINAL)
-	ZEND_NAMED_ME(__construct, PHP_FN(MongoDB_disabled___construct), ai_Symbol_void, ZEND_ACC_PRIVATE | ZEND_ACC_FINAL)
-	PHP_FE_END
-};
-/* clang-format on */
-/* }}} */
-
-/* {{{ MongoDB\BSON\Symbol object handlers */
+/* MongoDB\BSON\Symbol object handlers */
 static zend_object_handlers php_phongo_handler_symbol;
 
-static void php_phongo_symbol_free_object(zend_object* object) /* {{{ */
+static void php_phongo_symbol_free_object(zend_object* object)
 {
 	php_phongo_symbol_t* intern = Z_OBJ_SYMBOL(object);
 
@@ -254,9 +209,9 @@ static void php_phongo_symbol_free_object(zend_object* object) /* {{{ */
 		zend_hash_destroy(intern->properties);
 		FREE_HASHTABLE(intern->properties);
 	}
-} /* }}} */
+}
 
-zend_object* php_phongo_symbol_create_object(zend_class_entry* class_type) /* {{{ */
+zend_object* php_phongo_symbol_create_object(zend_class_entry* class_type)
 {
 	php_phongo_symbol_t* intern = zend_object_alloc(sizeof(php_phongo_symbol_t), class_type);
 
@@ -266,9 +221,9 @@ zend_object* php_phongo_symbol_create_object(zend_class_entry* class_type) /* {{
 	intern->std.handlers = &php_phongo_handler_symbol;
 
 	return &intern->std;
-} /* }}} */
+}
 
-static zend_object* php_phongo_symbol_clone_object(phongo_compat_object_handler_type* object) /* {{{ */
+static zend_object* php_phongo_symbol_clone_object(phongo_compat_object_handler_type* object)
 {
 	php_phongo_symbol_t* intern;
 	php_phongo_symbol_t* new_intern;
@@ -283,9 +238,9 @@ static zend_object* php_phongo_symbol_clone_object(phongo_compat_object_handler_
 	php_phongo_symbol_init(new_intern, intern->symbol, intern->symbol_len);
 
 	return new_object;
-} /* }}} */
+}
 
-static int php_phongo_symbol_compare_objects(zval* o1, zval* o2) /* {{{ */
+static int php_phongo_symbol_compare_objects(zval* o1, zval* o2)
 {
 	php_phongo_symbol_t *intern1, *intern2;
 
@@ -295,32 +250,23 @@ static int php_phongo_symbol_compare_objects(zval* o1, zval* o2) /* {{{ */
 	intern2 = Z_SYMBOL_OBJ_P(o2);
 
 	return strcmp(intern1->symbol, intern2->symbol);
-} /* }}} */
+}
 
-static HashTable* php_phongo_symbol_get_debug_info(phongo_compat_object_handler_type* object, int* is_temp) /* {{{ */
+static HashTable* php_phongo_symbol_get_debug_info(phongo_compat_object_handler_type* object, int* is_temp)
 {
 	*is_temp = 1;
 	return php_phongo_symbol_get_properties_hash(object, true);
-} /* }}} */
+}
 
-static HashTable* php_phongo_symbol_get_properties(phongo_compat_object_handler_type* object) /* {{{ */
+static HashTable* php_phongo_symbol_get_properties(phongo_compat_object_handler_type* object)
 {
 	return php_phongo_symbol_get_properties_hash(object, false);
-} /* }}} */
-/* }}} */
+}
 
-void php_phongo_symbol_init_ce(INIT_FUNC_ARGS) /* {{{ */
+void php_phongo_symbol_init_ce(INIT_FUNC_ARGS)
 {
-	zend_class_entry ce;
-
-	INIT_NS_CLASS_ENTRY(ce, "MongoDB\\BSON", "Symbol", php_phongo_symbol_me);
-	php_phongo_symbol_ce                = zend_register_internal_class(&ce);
+	php_phongo_symbol_ce                = register_class_MongoDB_BSON_Symbol(php_phongo_json_serializable_ce, php_phongo_type_ce, zend_ce_serializable);
 	php_phongo_symbol_ce->create_object = php_phongo_symbol_create_object;
-	PHONGO_CE_FINAL(php_phongo_symbol_ce);
-
-	zend_class_implements(php_phongo_symbol_ce, 1, php_phongo_json_serializable_ce);
-	zend_class_implements(php_phongo_symbol_ce, 1, php_phongo_type_ce);
-	zend_class_implements(php_phongo_symbol_ce, 1, zend_ce_serializable);
 
 #if PHP_VERSION_ID >= 80000
 	zend_class_implements(php_phongo_symbol_ce, 1, zend_ce_stringable);
@@ -333,4 +279,17 @@ void php_phongo_symbol_init_ce(INIT_FUNC_ARGS) /* {{{ */
 	php_phongo_handler_symbol.get_properties = php_phongo_symbol_get_properties;
 	php_phongo_handler_symbol.free_obj       = php_phongo_symbol_free_object;
 	php_phongo_handler_symbol.offset         = XtOffsetOf(php_phongo_symbol_t, std);
-} /* }}} */
+}
+
+bool phongo_symbol_new(zval* object, const char* symbol, size_t symbol_len)
+{
+	php_phongo_symbol_t* intern;
+
+	object_init_ex(object, php_phongo_symbol_ce);
+
+	intern             = Z_SYMBOL_OBJ_P(object);
+	intern->symbol     = estrndup(symbol, symbol_len);
+	intern->symbol_len = symbol_len;
+
+	return true;
+}
