@@ -71,12 +71,11 @@ mongoc_topology_reconcile (const mongoc_topology_t *topology,
 {
    mongoc_set_t *servers;
    mongoc_server_description_t *sd;
-   int i;
    mongoc_topology_scanner_node_t *ele, *tmp;
 
    servers = mc_tpld_servers (td);
    /* Add newly discovered nodes */
-   for (i = 0; i < (int) servers->items_len; i++) {
+   for (size_t i = 0u; i < servers->items_len; i++) {
       sd = mongoc_set_get_item (servers, i);
       _mongoc_topology_reconcile_add_nodes (sd, topology->scanner);
    }
@@ -212,6 +211,8 @@ _server_session_init (mongoc_server_session_t *session,
                       mongoc_topology_t *unused,
                       bson_error_t *error)
 {
+   BSON_UNUSED (unused);
+
    _mongoc_server_session_init (session, error);
 }
 
@@ -219,6 +220,8 @@ static void
 _server_session_destroy (mongoc_server_session_t *session,
                          mongoc_topology_t *unused)
 {
+   BSON_UNUSED (unused);
+
    _mongoc_server_session_destroy (session);
 }
 
@@ -226,10 +229,6 @@ static int
 _server_session_should_prune (mongoc_server_session_t *session,
                               mongoc_topology_t *topo)
 {
-   bool is_loadbalanced;
-   int timeout;
-   mc_shared_tpld td;
-
    BSON_ASSERT_PARAM (session);
    BSON_ASSERT_PARAM (topo);
 
@@ -244,9 +243,9 @@ _server_session_should_prune (mongoc_server_session_t *session,
    }
 
    /* Check for a timeout */
-   td = mc_tpld_take_ref (topo);
-   timeout = td.ptr->session_timeout_minutes;
-   is_loadbalanced = td.ptr->type == MONGOC_TOPOLOGY_LOAD_BALANCED;
+   mc_shared_tpld td = mc_tpld_take_ref (topo);
+   const int64_t timeout = td.ptr->session_timeout_minutes;
+   const bool is_loadbalanced = td.ptr->type == MONGOC_TOPOLOGY_LOAD_BALANCED;
    mc_tpld_drop_ref (&td);
 
    /** Load balanced topology sessions never expire */
@@ -267,32 +266,29 @@ _tpld_destroy_and_free (void *tpl_descr)
 
 const mongoc_host_list_t **
 _mongoc_apply_srv_max_hosts (const mongoc_host_list_t *hl,
-                             const int32_t max_hosts,
-                             size_t *const hl_array_size)
+                             size_t max_hosts,
+                             size_t *hl_array_size)
 {
-   size_t hl_size;
-   size_t idx;
    const mongoc_host_list_t **hl_array;
 
-   BSON_ASSERT (max_hosts >= 0);
    BSON_ASSERT_PARAM (hl_array_size);
 
-   hl_size = (size_t) _mongoc_host_list_length (hl);
+   const size_t hl_size = _mongoc_host_list_length (hl);
 
-   if (hl_size == 0) {
-      *hl_array_size = 0;
+   if (hl_size == 0u) {
+      *hl_array_size = 0u;
       return NULL;
    }
 
    hl_array = bson_malloc (hl_size * sizeof (mongoc_host_list_t *));
 
-   for (idx = 0u; hl; hl = hl->next) {
+   for (size_t idx = 0u; hl; hl = hl->next) {
       hl_array[idx++] = hl;
    }
 
-   if (max_hosts == 0 ||             /* Unlimited. */
-       hl_size == 1u ||              /* Trivial case. */
-       hl_size <= (size_t) max_hosts /* Already satisfies limit. */
+   if (max_hosts == 0u ||   // Unlimited.
+       hl_size == 1u ||     // Trivial case.
+       hl_size <= max_hosts // Already satisfies limit.
    ) {
       /* No random shuffle or selection required. */
       *hl_array_size = hl_size;
@@ -303,7 +299,7 @@ _mongoc_apply_srv_max_hosts (const mongoc_host_list_t *hl,
     * and less than the number of hosts in the DNS result, the driver MUST
     * randomly select that many hosts and use them to populate the seedlist.
     * Drivers SHOULD use the `Fisher-Yates shuffle` for randomization. */
-   for (idx = hl_size - 1u; idx > 0u; --idx) {
+   for (size_t idx = hl_size - 1u; idx > 0u; --idx) {
       /* 0 <= swap_pos <= idx */
       const size_t swap_pos =
          _mongoc_rand_size_t (0u, idx, _mongoc_simple_rand_size_t);
@@ -336,8 +332,6 @@ _mongoc_apply_srv_max_hosts (const mongoc_host_list_t *hl,
 mongoc_topology_t *
 mongoc_topology_new (const mongoc_uri_t *uri, bool single_threaded)
 {
-   int64_t heartbeat_default;
-   int64_t heartbeat;
    mongoc_topology_t *topology;
    mongoc_topology_description_type_t init_type;
    mongoc_topology_description_t *td;
@@ -366,15 +360,16 @@ mongoc_topology_new (const mongoc_uri_t *uri, bool single_threaded)
                                                   topology);
 
    topology->valid = false;
-   heartbeat_default =
+
+   const int32_t heartbeat_default =
       single_threaded ? MONGOC_TOPOLOGY_HEARTBEAT_FREQUENCY_MS_SINGLE_THREADED
                       : MONGOC_TOPOLOGY_HEARTBEAT_FREQUENCY_MS_MULTI_THREADED;
 
-   heartbeat = mongoc_uri_get_option_as_int32 (
+   const int32_t heartbeat = mongoc_uri_get_option_as_int32 (
       uri, MONGOC_URI_HEARTBEATFREQUENCYMS, heartbeat_default);
 
    topology->_shared_descr_._sptr_ = mongoc_shared_ptr_create (
-      bson_malloc0 (sizeof (mongoc_topology_description_t)),
+      BSON_ALIGNED_ALLOC0 (mongoc_topology_description_t),
       _tpld_destroy_and_free);
    td = mc_tpld_unsafe_get_mutable (topology);
    mongoc_topology_description_init (td, heartbeat);
@@ -519,6 +514,10 @@ mongoc_topology_new (const mongoc_uri_t *uri, bool single_threaded)
    td->max_hosts =
       mongoc_uri_get_option_as_int32 (uri, MONGOC_URI_SRVMAXHOSTS, 0);
 
+   if (td->max_hosts < 0) {
+      topology->valid = false;
+   }
+
    /*
     * Set topology type from URI:
     *   + if directConnection=true
@@ -592,24 +591,23 @@ mongoc_topology_new (const mongoc_uri_t *uri, bool single_threaded)
       return topology;
    }
 
-   {
-      size_t idx = 0u;
-      size_t hl_array_size = 0u;
+   size_t hl_array_size = 0u;
+
+   BSON_ASSERT (bson_in_range_signed (size_t, td->max_hosts));
+   const mongoc_host_list_t *const *hl_array =
+      _mongoc_apply_srv_max_hosts (hl, (size_t) td->max_hosts, &hl_array_size);
+
+   for (size_t idx = 0u; idx < hl_array_size; ++idx) {
+      const mongoc_host_list_t *const elem = hl_array[idx];
+
       uint32_t id = 0u;
 
-      const mongoc_host_list_t *const *hl_array =
-         _mongoc_apply_srv_max_hosts (hl, td->max_hosts, &hl_array_size);
-
-      for (idx = 0u; idx < hl_array_size; ++idx) {
-         const mongoc_host_list_t *const elem = hl_array[idx];
-
-         mongoc_topology_description_add_server (td, elem->host_and_port, &id);
-         mongoc_topology_scanner_add (topology->scanner, elem, id, false);
-      }
-
-      bson_free ((void *) hl_array);
+      mongoc_topology_description_add_server (td, elem->host_and_port, &id);
+      mongoc_topology_scanner_add (topology->scanner, elem, id, false);
    }
 
+   bson_free ((void *) hl_array);
+
    return topology;
 }
 
@@ -696,10 +694,14 @@ mongoc_topology_destroy (mongoc_topology_t *topology)
    mongoc_shared_ptr_reset_null (&topology->_shared_descr_._sptr_);
    mongoc_topology_scanner_destroy (topology->scanner);
    mongoc_server_session_pool_free (topology->session_pool);
+   bson_free (topology->clientSideEncryption.autoOptions.extraOptions
+                 .cryptSharedLibPath);
 
    mongoc_cond_destroy (&topology->cond_client);
    bson_mutex_destroy (&topology->tpld_modification_mtx);
 
+   bson_destroy (topology->encrypted_fields_map);
+
    bson_free (topology);
 }
 
@@ -930,9 +932,6 @@ mongoc_topology_compatible (const mongoc_topology_description_t *td,
                             const mongoc_read_prefs_t *read_prefs,
                             bson_error_t *error)
 {
-   int64_t max_staleness_seconds;
-   int32_t max_wire_version;
-
    if (td->compatibility_error.code) {
       if (error) {
          memcpy (error, &td->compatibility_error, sizeof (bson_error_t));
@@ -945,21 +944,10 @@ mongoc_topology_compatible (const mongoc_topology_description_t *td,
       return true;
    }
 
-   max_staleness_seconds =
+   const int64_t max_staleness_seconds =
       mongoc_read_prefs_get_max_staleness_seconds (read_prefs);
 
    if (max_staleness_seconds != MONGOC_NO_MAX_STALENESS) {
-      max_wire_version =
-         mongoc_topology_description_lowest_max_wire_version (td);
-
-      if (max_wire_version < WIRE_VERSION_MAX_STALENESS) {
-         bson_set_error (error,
-                         MONGOC_ERROR_COMMAND,
-                         MONGOC_ERROR_PROTOCOL_BAD_WIRE_VERSION,
-                         "Not all servers support maxStalenessSeconds");
-         return false;
-      }
-
       /* shouldn't happen if we've properly enforced wire version */
       if (!mongoc_topology_description_all_sds_have_write_date (td)) {
          bson_set_error (error,
@@ -1629,7 +1617,7 @@ _mongoc_topology_push_server_session (mongoc_topology_t *topology,
     * The next pop operation that encounters an expired session will clear the
     * entire session pool, thus preventing unbounded growth of the pool.
     */
-   mongoc_server_session_pool_return (server_session);
+   mongoc_server_session_pool_return (topology->session_pool, server_session);
 
    EXIT;
 }
@@ -1675,13 +1663,13 @@ _mongoc_topology_end_sessions_cmd (mongoc_topology_t *topology, bson_t *cmd)
       const char *key;
       bson_uint32_to_string (i, &key, buf, sizeof buf);
       BSON_APPEND_DOCUMENT (&ar, key, &ss->lsid);
-      mongoc_server_session_pool_drop (ss);
+      mongoc_server_session_pool_drop (topology->session_pool, ss);
    }
 
    if (ss) {
       /* We deleted at least 10'000 sessions, so we will need to return the
        * final session that we didn't drop */
-      mongoc_server_session_pool_return (ss);
+      mongoc_server_session_pool_return (topology->session_pool, ss);
    }
 
    bson_append_array_end (cmd, &ar);
@@ -1737,6 +1725,8 @@ _handle_sdam_app_error_command (mongoc_topology_t *topology,
    mc_tpld_modification tdmod;
    mongoc_server_description_t *mut_sd;
 
+   BSON_UNUSED (td);
+
    if (_mongoc_cmd_check_ok_no_wce (
           reply, MONGOC_ERROR_API_VERSION_2, &cmd_error)) {
       /* No error. */
@@ -1934,12 +1924,11 @@ _topology_collect_errors (const mongoc_topology_description_t *td,
 {
    const mongoc_server_description_t *server_description;
    bson_string_t *error_message;
-   int i;
 
    memset (error_out, 0, sizeof (bson_error_t));
    error_message = bson_string_new ("");
 
-   for (i = 0; i < mc_tpld_servers_const (td)->items_len; i++) {
+   for (size_t i = 0u; i < mc_tpld_servers_const (td)->items_len; i++) {
       const bson_error_t *error;
 
       server_description = mc_tpld_servers_const (td)->items[i].item;
@@ -2031,3 +2020,17 @@ mc_tpld_modify_drop (mc_tpld_modification mod)
    bson_mutex_unlock (&mod.topology->tpld_modification_mtx);
    mongoc_topology_description_destroy (mod.new_td);
 }
+
+bool
+mongoc_topology_uses_server_api (const mongoc_topology_t *topology)
+{
+   BSON_ASSERT_PARAM (topology);
+   return mongoc_topology_scanner_uses_server_api (topology->scanner);
+}
+
+bool
+mongoc_topology_uses_loadbalanced (const mongoc_topology_t *topology)
+{
+   BSON_ASSERT_PARAM (topology);
+   return mongoc_topology_scanner_uses_loadbalanced (topology->scanner);
+}
